from .base import Function
# from .nn_hyperparams import NNhyperparams
from .test_functions import three_hump_camel, rosenbrock, ackley, levi, himmelblau, tang, holder#, submanifold, nonlinear_submanifold, optimal_control

__all__ = [
    'Function',
    'three_hump_camel',
    'rosenbrock',
    'ackley',
    'levi',
    'himmelblau',
    'tang',
    'holder',
    # 'submanifold',
    # 'nonlinear_submanifold',
    # 'optimal_control'
    # 'NNhyperparams',
]


from .base import Surrogate

import numpy as np

class DGP(Surrogate):
    def __init__(self, **kwargs):
        super(DGP, self).__init__()

        self.name        = 'DGP'
        self.model       = kwargs['model'] if 'model' in kwargs else None
        self.hidden_size = kwargs['hidden_size'] if 'hidden_size' in kwargs else 64
        self.n_epochs    = kwargs['n_epochs'] if 'n_epochs' in kwargs else 100
        self.lr          = kwargs['lr'] if 'lr' in kwargs else 1e-3
        self.batch_size  = kwargs['batch_size'] if 'batch_size' in kwargs else 16
        self.verbose     = kwargs['verbose'] if 'verbose' in kwargs else 1
    
    def fit(self, X, y):

        if self.model is None:
            self.model = DistributionalDGP(X.shape[-1], y.shape[-1],
                                           n_epochs=self.n_epochs, hidden_size=self.hidden_size,
                                           lr = self.lr, batch_size=self.batch_size, verbose=self.verbose)
        
        self.model.fit(X, y)
    
    def predict(self, X):
        # print("predict in DGP(Surrogate): ", torch.tensor(X).dtype)

        f, std = self.model.predict(torch.tensor(X))

        return f, np.sqrt(std)


import torch
from torch.utils.data import TensorDataset, DataLoader
# from wgpot import Wasserstein_GP
from tqdm import tqdm
import gpytorch
from gpytorch.means import ConstantMean, LinearMean
from gpytorch.kernels import RBFKernel, ScaleKernel
from gpytorch.variational import VariationalStrategy, CholeskyVariationalDistribution
from gpytorch.distributions import MultivariateNormal
from gpytorch.mlls import VariationalELBO
from gpytorch.likelihoods import GaussianLikelihood

from gpytorch.models.deep_gps import DeepGPLayer, DeepGP
from gpytorch.mlls import DeepApproximateMLL
import gpytorch.mlls
import numpy as np
    

num_hidden_dims = 2

class ToyDeepGPHiddenLayer2(DeepGPLayer):
    def __init__(self, input_dims, output_dims, num_inducing=128, mean_type='constant'):
        if output_dims is None:
            inducing_points = torch.randn(num_inducing, input_dims)
            batch_shape = torch.Size([])
        else:
            inducing_points = torch.randn(output_dims, num_inducing, input_dims)
            batch_shape = torch.Size([output_dims])

        variational_distribution = CholeskyVariationalDistribution(
            num_inducing_points=num_inducing,
            batch_shape=batch_shape
        )

        variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True
        )

        super(ToyDeepGPHiddenLayer2, self).__init__(variational_strategy, input_dims, output_dims)

        if mean_type == 'constant':
            self.mean_module = ConstantMean(batch_shape=batch_shape)
        else:
            self.mean_module = LinearMean(input_dims)
        self.covar_module = ScaleKernel(
            RBFKernel(batch_shape=batch_shape, ard_num_dims=input_dims),
            batch_shape=batch_shape, ard_num_dims=None
        )

    def forward(self, x):
        mean_x = self.mean_module(x.float())
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)

    def __call__(self, x, *other_inputs, **kwargs):
        """
        Overriding __call__ isn't strictly necessary, but it lets us add concatenation based skip connections
        easily. For example, hidden_layer2(hidden_layer1_outputs, inputs) will pass the concatenation of the first
        hidden layer's outputs and the input data to hidden_layer2.
        """
        # if len(other_inputs):
        #     if isinstance(x, gpytorch.distributions.MultitaskMultivariateNormal):
        #         x = x.rsample()

        #     processed_inputs = [
        #         inp.unsqueeze(0).expand(gpytorch.settings.num_likelihood_samples.value(), *inp.shape)
        #         for inp in other_inputs
        #     ]

        #     x = torch.cat([x] + processed_inputs, dim=-1)

        return super().__call__(x)


class DistributionalDGP(DeepGP):
    def __init__(self, input_dim, out_dim, **kwargs):
        super().__init__()
        hidden_layer = ToyDeepGPHiddenLayer2(
            input_dims=input_dim,
            output_dims=kwargs.get('hidden_size', 16),
            mean_type='linear',
        )

        last_layer = ToyDeepGPHiddenLayer2(
            input_dims=hidden_layer.output_dims,
            output_dims=None,
            mean_type='linear', # TODO LINEAR
        )

        self.hidden_layer = hidden_layer
        self.last_layer = last_layer
        self.likelihood = GaussianLikelihood()
        self.n_epochs = kwargs.get('n_epochs', 100)
        self.num_samples = kwargs.get('num_samples', 100)
        self.lr = kwargs.get('lr', 1e-3)
        self.batch_size = kwargs.get('batch_size', 8)
        self.verbose = kwargs.get('verbose', 1)

    def fit(self, X, y):
      
      self.loss_history = []
      self.predictions_history = []

      optimizer = torch.optim.Adam([{'params': self.parameters()},], lr=self.lr)
      mll = DeepApproximateMLL(VariationalELBO(self.likelihood, self, 500))


      epochs_iter = tqdm(range(self.n_epochs), desc="Epoch") if self.verbose > 1 else range(self.n_epochs)

    #   print("Fit in class Distributional GP: ", X.dtype)  
      for _ in epochs_iter:
        closs = 0
        with gpytorch.settings.num_likelihood_samples(self.num_samples):
            for x_batch, y_batch in DataLoader(TensorDataset(torch.tensor(X, dtype=torch.float32), torch.tensor(y, dtype=torch.float32)), batch_size=self.batch_size, shuffle=True):
                # print("Fit in class Distributional GP x_batch: ", x_batch.dtype)  
                output = self(x_batch)
                loss = -mll(output, y_batch)
                loss.backward(retain_graph=True)
                closs += loss.item()
                optimizer.step()

                optimizer.zero_grad()
        
        if self.verbose > 1:
            epochs_iter.set_description(f"Loss: {closs:.3f}")
        self.loss_history.append(loss)

        # with torch.no_grad():
        #     self.predictions_history.append(self.forward(train_mean_x))
      return self.loss_history

    def forward(self, x):
        with gpytorch.settings.num_likelihood_samples(1):
            hidden_rep1 = self.hidden_layer(x)
            output = self.last_layer(hidden_rep1)
        # mean_x = self.mean_module(output)
        # covar_x = self.covar_module(x)
        return output
    
    def predict(self, x):
        with torch.no_grad():
            mus = []
            variances = []
            # lls = []
            # for x_batch, y_batch in test_loader:
            # print("Distributional DGP predict; ", x.float().dtype)
            preds = self.likelihood(self(x.float()))
            mus.append(preds.mean)
            variances.append(preds.variance)
            # lls.append(self.likelihood.log_marginal(y, self(x)))

        return torch.cat(mus, dim=-1), torch.cat(variances, dim=-1)
    
    def __call__(self, x, *other_inputs, **kwargs):
        """
        Overriding __call__ isn't strictly necessary, but it lets us add concatenation based skip connections
        easily. For example, hidden_layer2(hidden_layer1_outputs, inputs) will pass the concatenation of the first
        hidden layer's outputs and the input data to hidden_layer2.
        """
        with gpytorch.settings.num_likelihood_samples(self.num_samples):
            if len(other_inputs):
                if isinstance(x, gpytorch.distributions.MultitaskMultivariateNormal):
                    x = x.rsample()

                processed_inputs = [
                    inp.unsqueeze(0).expand(gpytorch.settings.num_likelihood_samples.value(), *inp.shape)
                    for inp in other_inputs
                ]

                x = torch.cat([x] + processed_inputs, dim=-1)

            # Метод call идет из базового класса Module, по умолчанию передает содержимое в forward, затем проверяет, чтобы тип выходной переменной был один из этих - (Distribution, torch.Tensor, LinearOperator)
            # Принимает значения размерности от двух
        return super().__call__(x)


import argparse
from waggon.utils import display
from waggon import functions as f
from waggon.optim import SurrogateOptimiser
from waggon.acquisitions import WU, WU_IDW, EI, CB
from waggon.surrogates import BNN, DE, DGP, GAN, GP

FUNCS = {
    'ackley':     f.ackley,
    'himmelblau': f.himmelblau,
    'holder':     f.holder,
    'levi':       f.levi,
    'rosenbrock': f.rosenbrock,
    'tang':       f.tang,
    'thc':        f.three_hump_camel
}

SURR = {
    'bnn' : BNN(),
    'de' :  DE(),
    'dgp' : DGP(),
    'gan' : GAN(),
    'gp' :  GP()
}

ACQF = {
    'wu':  WU(),
    'wu-idw': WU_IDW(),
    'ei':  EI(),
    'lcb': CB(minimise=True),
    'ucb': CB(minimise=False)
}

SEEDS = [2, 3, 5, 7, 11, 13, 17, 19, 23, 73]

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--function', help='optimise the given function', default='thc', choices=['thc', 'ackley', 'levi', 'himmelblau', 'rosenbrock', 'tang', 'holder'])
    parser.add_argument('-d', '--dimensions', type=int, help='dimensionality of the experiment', default=None)
    parser.add_argument('-s', '--surrogate', help='surrogate for optimisation', default='gan', choices=['gan', 'bnn', 'de', 'dgp', 'gp'])
    parser.add_argument('-a', '--acquisition', help='acqusition function to use for optimisation', default='wu', choices=['wu', 'wu-idw', 'ei', 'lcb', 'ucb'])
    parser.add_argument('-v', '--verbose', type=int, help='increase output verbose', choices=[0, 1, 2], default=1)

    args = parser.parse_args()
    
    if (args.surrogate == 'gan' and not ('wu' in args.acquisition)) or (args.surrogate != 'gan' and 'wu' in args.acquisition):
        raise ValueError(f'Surrogate {args.surrogate} is not compatible with {args.acquisition} acquisition function')

    for i, seed in enumerate(SEEDS):
        
        print(f'Experiment #{i}')

        opt = SurrogateOptimiser(func=FUNCS[args.function](args.dimensions) if args.dimensions else FUNCS[args.function](),
                                surr=SURR[args.surrogate],
                                acqf=ACQF[args.acquisition],
                                verbose=args.verbose,
                                seed=seed,
                                save_results=True)
        opt.optimise()
        display(opt)

if __name__ == '__main__':
    main()



import argparse

import waggon.functions as f
from waggon.optim import DifferentialEvolutionOptimizer

FUNCS = {
    'ackley':     f.ackley,
    'himmelblau': f.himmelblau,
    'holder':     f.holder,
    'levi':       f.levi,
    'rosenbrock': f.rosenbrock,
    'tang':       f.tang,
    'thc':        f.three_hump_camel
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--function', help='Function to optimize', default='levi', choices=['thc', 'ackley', 'levi', 'himmelblau', 'rosenbrock', 'tang', 'holder']
    )
    parser.add_argument(
        '-d', '--dimensions', type=int, help='Dimensionality of the experiment', default=None
    )
    parser.add_argument(
        '-v', '--verbose', type=int, help='increase output verbose', choices=[0, 1, 2], default=1
    )
    parser.add_argument(
        '-e', '--experiments', help="Number of experiments", default=5, type=int
    )
    parser.add_argument(
        '-eps', '--epsilon', help="Required error rate", default=1e-1, type=float
    )
    parser.add_argument(
        '-n', '--n_candidates', help="Number of points in generation", type=int, default=11 ** 2
    )

    args = parser.parse_args()

    for i in range(args.experiments):
        print(f"Experiment #{i}")

        func = FUNCS[args.function]
        dim = args.dimensions

        if dim is not None:
            func = func(dim)
        else:
            func = func()
        
        eps = args.epsilon
        n_candidates = args.n_candidates
        
        evolution = DifferentialEvolutionOptimizer(
            func=func, eps=eps, n_candidates=n_candidates
        )
        evolution.optimise()


if __name__ == "__main__":
    main()



import numpy as np
import jax.numpy as jnp
from protes import protes

from ..base import Optimiser


class ProtesOptimiser(Optimiser):
    def __init__(self, func, K=10, constraint_func=None, penalty=1e6, **kwargs):
        """
        Initialize the PROTES Optimiser.

        Parameters:
        - func: The function to optimize.
        - K: Number of samples per iteration.
        - constraint_func: Function to evaluate constraints on candidates.
        - penalty: Penalty value for constraint violations.
        """
        super().__init__(**kwargs)
        self.func = func
        self.K = K
        self.constraint_func = constraint_func
        self.penalty = penalty
        self.verbose = kwargs.get("verbose", True)

        self.d = self.func.T
        self.n = 2  
        self.protes_model = self.initialize_protes()

    def initialize_protes(self):
        if self.verbose:
            print(f"Initializing PROTES with d={self.d}, n={self.n}, K={self.K}")
        
        return lambda f: protes(f=f, d=self.d, n=self.n, m=self.K, log=self.verbose)

    def f_batch(self, batch):
        func_values = jnp.array([self.func(x) for x in batch])

        if self.constraint_func is not None:
            penalties = jnp.array([self.constraint_func(x) for x in batch])
            func_values += penalties * self.penalty

        return func_values

    def predict(self):

        f_batch = lambda batch: self.f_batch(jnp.array(batch))
        try:
            best_candidate, best_value = self.protes_model(f=f_batch)
        except Exception as e:
            print(f"Error during PROTES optimization: {e}")
            raise

        if self.verbose:
            print(f"Best candidate: {best_candidate}, Best value: {best_value}")

        return best_candidate, best_value



import time
import pickle
import numpy as np
from scipy.optimize import minimize

from .base import Optimiser
from .utils import create_dir

class SurrogateOptimiser(Optimiser):
    '''
    Surrogate based black-box optimiser.

    Parameter
    ----------
    func : Callable, waggon.functions.Function
        Black-box function to be optimised.
    
    surr : waggon.surrogates.Surrogate or waggon.surrogate.GenSurrogate
        Surrogate model for the black-box function.
    
    acqf : waggon.acquisitions.Acquisition
        Acquisition function defining the optimisation strategy.
    
    max_iter : int, default = 100
        Maximum number of optimisation loop iterations.
    
    eps : float, default = 1e-1
        Epsilon-solution criterion value.
    
    error_type : {'x', 'f'}, default = 'x'
        Optimisation error type - either in argument, 'x', or function, 'f'.
    
    num_opt : bool, default = False
        Whether the acquisition function is optimised numerically or not. If False,
        the search of the next best parameters is done via direct search.
    
    eq_cons : dict, default = None
        Equality-type constraints for numerical optimisation.
    
    ineq_cons : dict, default = None
        Ineqaulity-type constraints for numerical optimisation.

    fix_candidates : bool, default = False if num_opt else True
        Whether the candidate points should be fixed or not.
    
    n_candidates : int, default = 1 if num_opt else 101**2
        Number of candidates points.
    
    olhs : bool, default = True
        Whether orthogonal Latin hypercube sampling (LHS) is used for choosing candidate points.
        If False, simple LHS is used.

    lhs_seed : int, default = None
        Controls the randomness of candidates sampled via (orthogonal) LHS.
    
    verbose : int, default = 1
        Controls verbosity when fitting and predicting. By default only a progress bar over the
        optimisation loop is displayed.
    
    save_results : bool, default = True
        Whether results are saved or not.
    '''
    def __init__(self, func, surr, acqf, **kwargs):
        super(SurrogateOptimiser).__init__()
        
        self.func           = func
        self.surr           = surr
        self.acqf           = acqf
        self.max_iter       = kwargs['max_iter'] if 'max_iter' in kwargs else 100
        self.eps            = kwargs['eps'] if 'eps' in kwargs else 1e-1
        self.error_type     = kwargs['error_type'] if 'error_type' in kwargs else 'x'
        self.num_opt        = kwargs['num_opt'] if 'num_opt' in kwargs else False
        self.eq_cons        = kwargs['eq_cons'] if 'eq_cons' in kwargs else None
        self.ineq_cons      = kwargs['ineq_cons'] if 'ineq_cons' in kwargs else None
        self.fix_candidates = False if self.num_opt else (kwargs['fix_candidates'] if 'fix_candidates' in kwargs else True)
        self.n_candidates   = kwargs['n_candidates'] if 'n_candidates' in kwargs else (1 if self.num_opt else 101**2)
        self.olhs           = kwargs['olhs'] if 'olhs' in kwargs else True
        self.lhs_seed       = kwargs['lhs_seed'] if 'lhs_seed' in kwargs else None
        self.verbose        = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.save_results   = kwargs['save_results'] if 'save_results' in kwargs else True
        self.surr.verbose   = self.verbose
        self.candidates     = None

    def numerical_search(self):
        '''
        Numerical optimisation of the acquisition function.

        Returns
        -------
        best_x : np.array of shape (func.dim,)
            Predicted optimum of the acquisition function.
        '''
        best_x = None
        best_acqf = np.inf

        candidates = self.create_candidates()

        for x0 in candidates:

            if (self.eq_cons is None) and (self.ineq_cons is None):
                opt_res = minimize(method='L-BFGS-B', fun=self.acqf, x0=x0, bounds=self.func.domain)
            else:
                opt_res = minimize(method='SLSQP', fun=self.acqf, x0=x0, bounds=self.func.domain, constraints=[self.eq_cons, self.ineq_cons])
            
            if opt_res.fun < best_acqf:
                best_acqf = opt_res.fun
                best_x = opt_res.x
        
        return best_x
    
    def direct_search(self):
        '''
        Direct search for the optimum of the acquisition function.

        Returns
        -------
        best_x : np.array of shape (func.dim,)
            Predicted optimum of the acquisition function
        '''
        
        if self.fix_candidates:
            if self.candidates is None:
                self.candidates = self.create_candidates()
        else:
            self.candidates = self.create_candidates()
        
        if self.eq_cons is not None:
            self.candidates = self.candidates[np.where(np.all(self.eq_cons(self.candidates) == 0, axis=0))[0]]
        if self.ineq_cons is not None:
            self.candidates = self.candidates[np.where(np.all(self.eq_cons(self.candidates) <= 0, axis=0))[0]]
        
        acqf_values = self.acqf(self.candidates)
        best_x = self.candidates[np.argmin(acqf_values)]
        
        return best_x

    def predict(self, X, y, n_pred=1):
        '''
        Predicts the next best set of parameter values by optimising the acquisition function.
        
        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Training input points.

        y : np.array of shape (n_samples * func.n_obs, func.dim)
            Target values of the black-box function.
        
        n_pred : int, defult = 1
            Number of predictions to make.
        
        Returns
        -------
        next_x : np.array of shape (func.dim, n_pred)
        '''

        self.surr.fit(X, y)
            
        self.acqf.y = y.reshape(y.shape[0]//self.func.n_obs, self.func.n_obs)
        self.acqf.conds = X[::self.func.n_obs]
        self.acqf.surr = self.surr

        next_x = []

        for j in range(n_pred):
            if self.num_opt:
                next_x.append(self.numerical_search())
            else:
                next_x.append(self.direct_search())

        return np.array(next_x)
    
    def _save(self, base_dir='test_results'):
        res_path = create_dir(self.func, self.acqf.name, self.surr.name, base_dir=base_dir)

        with open(f'{res_path}/{time.strftime("%d_%m_%H_%M_%S")}.pkl', 'wb') as f:
            pickle.dump(self.res, f)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="waggon",
    version="0.1.7",
    description="WAsserstein Global Gradient-free OptimisatioN (WAGGON) methods library.",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hse-cs/waggon",
    author="Tigran Ramazyan",
    author_email="ramazyant@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    install_requires=["GPy==1.13.2", "gpytorch==1.10.0", "matplotlib==3.7.5", "numpy==1.26.0",
                      "scikit_learn==1.3.2", "scipy==1.12.0", "setuptools==72.1.0",
                      "torch==2.2.2", "torchbnn==1.2", "torchensemble==0.1.9", "tqdm==4.66.4", "jax==0.5.0", "protes==0.3.11"],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.11",
)



from .base import Optimiser
from .surrogate import SurrogateOptimiser
from .baselines.evolutions import DifferentialEvolutionOptimizer
from .baselines.protes import ProtesOptimiser

__all__ = [
    'Optimiser',
    'SurrogateOptimiser',
    "DifferentialEvolutionOptimizer",
    'ProtesOptimiser'
]



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch.utils.data import TensorDataset, DataLoader

from sklearn.datasets import load_iris # change to MNIST?
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split

import numpy as np

from .base import Function


class MLP(nn.Module):
    def __init__(self, input_dim, search_space, **kwargs):
        super(MLP,self).__init__()

        self.input_layer    = nn.Linear(input_dim, search_space['hidden_size0'])
        self.hidden_layer1  = nn.Linear(search_space['hidden_size0'], search_space['hidden_size1'])
        self.output_layer   = nn.Linear(search_space['hidden_size1'], 3)

        self.learning_rate = kwargs['learning_rate'] if 'learning_rate' in kwargs else 1e-2
        self.criterion = kwargs['criterion'] if 'criterion' in kwargs else nn.CrossEntropyLoss()
        self.optimiser = kwargs['optimiser'] if 'optimiser' in kwargs else torch.optim.Adam(self.parameters(),lr=self.learning_rate)
        self.n_epochs = kwargs['n_epochs'] if 'n_epochs' in kwargs else 100
        self.batch_size = kwargs['batch_size'] if 'batch_size' in kwargs else 128

        self.score = kwargs['score'] if 'score' in kwargs else accuracy_score

    def forward(self, x):
        x = F.relu(self.input_layer(x))
        x = F.relu(self.hidden_layer1(x))
        x = F.softmax(self.output_layer(x), dim=-1)
        return x
    
    def fit(self, X_train, y_train):

        dataset = TensorDataset(X_train, y_train)

        for _ in range(self.n_epochs):

            for X_batch, y_batch in DataLoader(dataset, batch_size=self.batch_size, shuffle=True, drop_last=True):
            
                y_pred = self(X_batch)
                loss = self.criterion(y_pred, y_batch)
                
                self.optimiser.zero_grad()
                loss.backward()
                self.optimiser.step()

    def val(self, x_test, y_test):
        pred = self(x_test)
        pred = pred.detach().numpy()
        return self.score(y_test, np.argmax(pred, axis=1))


class NNhyperparams(Function):
    def __init__(self, model, **kwargs):

        self.dim           = kwargs['dim'] if 'dim' in kwargs else 2
        self.domain        = kwargs['domain'] if 'domain' in kwargs else np.array([[1, 100], [1, 100]])
        self.name          = 'NNhyperparams'
        self.glob_min      = np.array([120, 45])
        self.f             = lambda x: self.__call__(x)
        self.log_transform = kwargs['log_transform'] if 'log_transform' in kwargs else True
        self.log_eps       = kwargs['log_eps'] if 'log_eps' in kwargs else 1e-8
        self.sigma         = kwargs['sigma'] if 'sigma' in kwargs else 1e-1
        self.n_obs         = 13 * 3
        self.model         = model
        self.dataset       = kwargs['dataset'] if 'dataset' in kwargs else load_iris(return_X_y=True)
        self.minimise      = kwargs['minimise'] if 'minimise' in kwargs else True
        self.seed          = kwargs['seed'] if 'seed' in kwargs else 73

        self.X = Variable(torch.from_numpy(self.dataset[0])).float()
        self.y = Variable(torch.from_numpy(self.dataset[1]))
    

    def __call__(self, neurons : np.array):

        x_train, x_test, y_train, y_test = train_test_split(self.X, self.y, random_state=self.seed, shuffle=True)
        outputs = []

        for pack in neurons:

            self.model.__init__(x_train.shape[-1], pack)

            self.model.fit(x_train, y_train)
            roc_auc_ = self.model.val(x_test, y_test)
            outputs.append(roc_auc_)

        if self.minimise:
            return -1.0 * np.array(outputs)
        else:
            return np.array(outputs)
    

    def sample(self, neurons):

        X, y = None, None

        for n in neurons:
            X_ = np.array(self.n_obs * [n])

            x = torch.normal(torch.mean(self.X), torch.std(self.X), (self.n_obs, self.X.shape[-1]))
            proba = self.model(x).detach().numpy()
            y_ = self.f([n]) + proba

            if X is None:
                X, y = X_, y_
            else:
                X = np.concatenate((X, X_))
                y = np.concatenate((y, y_))

        return X, y
